# tests/test_retry_async.py
import asyncio
import inspect

import httpx
import pytest

import rediskit.retry_decorator as rm


@pytest.fixture
def sleep_calls(monkeypatch):
    """
    Replace asyncio.sleep with a fake that records requested sleep times (without actually sleeping).
    Returns a list you can assert on.
    """
    calls = []

    async def fake_sleep(s: float):
        calls.append(s)
        # do not actually sleep

    monkeypatch.setattr(asyncio, "sleep", fake_sleep)
    return calls


@pytest.fixture
def fixed_random(monkeypatch):
    """
    Deterministic RNG. By default yields 0.5 repeatedly -> zero jitter in your formula.
    Call fixed_random.set([...]) to change the sequence per-test.
    """
    seq = {"vals": [0.5]}

    def iterator():
        while True:
            for v in seq["vals"]:
                yield v

    it = iterator()

    def fake_random():
        return next(it)

    # Patch the random imported inside your module
    monkeypatch.setattr(rm.random, "random", fake_random)

    class Ctl:
        def set(self, values):
            seq["vals"] = values

    return Ctl()


def failing_then_success(n_fail: int, exc: Exception):
    """
    Build an async function that fails n_fail times with `exc`, then returns "OK".
    Also captures call count via attribute for assertions.
    """
    state = {"count": 0}

    @rm.retry_async()
    async def fn(*args, **kwargs):
        state["count"] += 1
        if state["count"] <= n_fail:
            raise exc
        return "OK"

    fn._state = state  # type: ignore # attach for test introspection (ok in tests)
    return fn


@pytest.mark.asyncio
async def test_success_without_retry():
    @rm.retry_async()
    async def fn():
        return 42

    assert await fn() == 42


@pytest.mark.asyncio
async def test_disabled_retry_does_not_retry(sleep_calls):
    # enabled=False should bypass retry and not sleep
    err = httpx.RequestError("boom")
    calls = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(enabled=False, attempts=5))
    async def fn():
        calls["n"] += 1
        raise err

    with pytest.raises(httpx.RequestError):
        await fn()

    assert calls["n"] == 1
    assert sleep_calls == []


@pytest.mark.asyncio
async def test_attempts_works_and_stops_on_success(sleep_calls, fixed_random):
    # 2 failures + 1 success with attempts=3
    fn = failing_then_success(2, httpx.RequestError("x"))

    result = await fn()
    assert result == "OK"
    # backoff default backoff=0.5, jitter ±0.1 but fixed_random produces 0 jitter
    assert 0.4 < sleep_calls[0] < 0.6 and 0.9 < sleep_calls[1] < 1.1  # exponential: 0.5, 1.0 for first two retries
    assert fn._state["count"] == 3


@pytest.mark.asyncio
async def test_exhausts_and_raises_after_max_attempts(sleep_calls, fixed_random):
    # fails 3 times, attempts=3 -> should raise after two sleeps
    err = httpx.RequestError("nope")
    state = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(attempts=3, backoff=0.2, jitter=0.0))
    async def fn():
        state["n"] += 1
        raise err

    with pytest.raises(httpx.RequestError):
        await fn()

    assert state["n"] == 3
    assert sleep_calls == [0.2, 0.4]


@pytest.mark.asyncio
async def test_non_retriable_exception_bubbles_no_sleep(sleep_calls):
    # ValueError not in default exceptions -> no retry
    state = {"n": 0}

    @rm.retry_async()
    async def fn():
        state["n"] += 1
        raise ValueError("bad")

    with pytest.raises(ValueError):
        await fn()

    assert state["n"] == 1
    assert sleep_calls == []


@pytest.mark.asyncio
async def test_cancelled_error_propagates_without_retry(sleep_calls):
    state = {"n": 0}

    @rm.retry_async()
    async def fn():
        state["n"] += 1
        raise asyncio.CancelledError()

    with pytest.raises(asyncio.CancelledError):
        await fn()

    assert state["n"] == 1
    assert sleep_calls == []


@pytest.mark.asyncio
async def test_http_status_error_is_retried(sleep_calls, fixed_random):
    # httpx.HTTPStatusError is in default tuple -> should retry
    req = httpx.Request("GET", "https://x")
    resp = httpx.Response(500, request=req)
    err = httpx.HTTPStatusError("server err", request=req, response=resp)

    state = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(attempts=2, backoff=0.1, jitter=0.0))
    async def fn():
        state["n"] += 1
        raise err

    with pytest.raises(httpx.HTTPStatusError):
        await fn()

    assert state["n"] == 2
    assert sleep_calls == [0.1]


@pytest.mark.asyncio
async def test_per_call_policy_override_with_dict(sleep_calls, fixed_random):
    # Override attempts/backoff/jitter/exceptions via dict
    req = httpx.Request("GET", "https://x")
    resp = httpx.Response(503, request=req)
    err = httpx.HTTPStatusError("svc err", request=req, response=resp)

    calls = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(attempts=5, backoff=1.0, jitter=0.0))
    async def fn():
        calls["n"] += 1
        raise err

    # override to attempts=2, backoff=0.2
    with pytest.raises(httpx.HTTPStatusError):
        await fn(retry_policy={"attempts": 2, "backoff": 0.2, "jitter": 0.0})

    assert calls["n"] == 2
    assert sleep_calls == [0.2]


@pytest.mark.asyncio
async def test_per_call_policy_override_with_object(sleep_calls, fixed_random):
    err = httpx.RequestError("oh no")
    calls = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(attempts=5, backoff=1.0, jitter=0.0))
    async def fn():
        calls["n"] += 1
        raise err

    policy = rm.RetryPolicy(attempts=3, backoff=0.05, jitter=0.0)
    with pytest.raises(httpx.RequestError):
        await fn(retry_policy=policy)

    assert calls["n"] == 3
    assert sleep_calls == [0.05, 0.10]


@pytest.mark.asyncio
async def test_retry_policy_kwarg_not_forwarded_to_fn_arguments(sleep_calls):
    # Ensure the decorator pops 'retry_policy' and doesn't pass it into the wrapped function
    seen_kwargs = {}

    @rm.retry_async(rm.RetryPolicy(attempts=1))
    async def fn(**kwargs):
        nonlocal seen_kwargs
        seen_kwargs = kwargs
        return "ok"

    res = await fn(retry_policy={"attempts": 1}, x=1)
    assert res == "ok"
    assert "retry_policy" not in seen_kwargs
    assert seen_kwargs == {"x": 1}


@pytest.mark.asyncio
async def test_attempts_equal_one_means_no_sleep_no_retry(sleep_calls):
    calls = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(attempts=1, backoff=0.5))
    async def fn():
        calls["n"] += 1
        raise httpx.RequestError("fail once")

    with pytest.raises(httpx.RequestError):
        await fn()

    assert calls["n"] == 1
    assert sleep_calls == []


@pytest.mark.asyncio
async def test_non_positive_backoff_still_behaves_and_never_sleeps_negative(sleep_calls, fixed_random):
    # backoff <= 0 should not cause negative sleeps; first delay is <=0 -> sleep(0)
    calls = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(attempts=3, backoff=-1.0, jitter=0.0))
    async def fn():
        calls["n"] += 1
        raise httpx.RequestError("x")

    with pytest.raises(httpx.RequestError):
        await fn()

    # delays: max(0, -1) -> 0, then doubles -> 0, then doubles -> 0 (but only two sleeps)
    assert sleep_calls == [0.0, 0.0]
    assert calls["n"] == 3


def test_preserves_function_metadata():
    async def original(a, b=2):
        """Docstring here."""
        return a + b

    wrapped = rm.retry_async()(original)

    assert wrapped.__name__ == "original"
    assert wrapped.__doc__ == "Docstring here."
    _sig_orig = inspect.signature(original)
    sig_wrap = inspect.signature(wrapped)
    # wrapper(*args, **kwargs) can't replicate exact signature at runtime,
    # but we at least ensure it's still a coroutine function and callable with same params.
    assert asyncio.iscoroutinefunction(wrapped)
    # quick callability check via param names
    for name in ["a", "b"]:
        assert name in str(sig_wrap)


@pytest.mark.asyncio
async def test_custom_exceptions_tuple(sleep_calls):
    # If exceptions tuple excludes httpx.RequestError, there should be no retry
    calls = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(attempts=3, exceptions=(httpx.HTTPStatusError,)))
    async def fn():
        calls["n"] += 1
        raise httpx.RequestError("no retry for this")

    with pytest.raises(httpx.RequestError):
        await fn()

    assert calls["n"] == 1
    assert sleep_calls == []


@pytest.mark.asyncio
async def test_success_after_first_retry_returns_value(sleep_calls, fixed_random):
    calls = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(attempts=4, backoff=0.05, jitter=0.0))
    async def fn(x):
        calls["n"] += 1
        if calls["n"] == 1:
            raise httpx.RequestError("transient")
        return x * 2

    val = await fn(7)
    assert val == 14
    assert calls["n"] == 2
    assert sleep_calls == [0.05]


@pytest.mark.asyncio
async def test_policy_kwarg_wrong_type_ignored_uses_default(sleep_calls, fixed_random):
    # Passing a nonsense retry_policy value should fall back to default policy used in decorator
    calls = {"n": 0}

    @rm.retry_async(rm.RetryPolicy(attempts=2, backoff=0.1, jitter=0.0))
    async def fn():
        calls["n"] += 1
        raise httpx.RequestError("x")

    with pytest.raises(httpx.RequestError):
        await fn(retry_policy=object())  # not dict/rm.RetryPolicy

    assert calls["n"] == 2
    assert sleep_calls == [0.1]



import asyncio
import importlib
import threading
from queue import Queue

import pytest

import rediskit.redis.a_client.redis_in_eventloop as loop_redis


class FakePool:
    def __init__(self):
        self.disconnected = False

    async def disconnect(self, inuse_connections=True):
        self.disconnected = True


class FakeClient:
    _instances = 0

    def __init__(self):
        FakeClient._instances += 1
        self.instance_no = FakeClient._instances
        self.loop = asyncio.get_running_loop()
        self.closed = False
        self.connection_pool = FakePool()

    async def ping(self):
        return True

    async def close(self):
        self.closed = True

    async def aclose(self):
        self.closed = True


@pytest.fixture(autouse=True)
def patch_factory_and_reset(monkeypatch):
    """
    - Patch _make_client to return a FakeClient bound to the *current* loop.
    - Reload the module between tests to reset the internal registry.
    """
    importlib.reload(loop_redis)
    factory_calls = {"n": 0}

    def _fake_factory():
        factory_calls["n"] += 1
        return FakeClient()

    monkeypatch.setattr(loop_redis, "_make_client", _fake_factory, raising=True)
    yield factory_calls  # let tests assert how many times we constructed clients


@pytest.mark.asyncio
async def test_same_loop_returns_same_client(patch_factory_and_reset):
    c1 = await loop_redis.get_async_redis_connection_in_eventloop()
    c2 = await loop_redis.get_async_redis_connection_in_eventloop()
    assert c1 is c2
    assert c1.loop is asyncio.get_running_loop()
    # Only one factory call for this loop
    assert patch_factory_and_reset["n"] == 1


@pytest.mark.asyncio
async def test_concurrent_first_use_is_single_init(patch_factory_and_reset):
    async def grab():
        return await loop_redis.get_async_redis_connection_in_eventloop()

    # many concurrent callers on first use
    results = await asyncio.gather(*[grab() for _ in range(50)])
    first = results[0]
    assert all(r is first for r in results)
    # still only one construction on this loop
    assert patch_factory_and_reset["n"] == 1


def run_coro_in_new_loop(coro) -> object:
    """
    Run a coroutine in a fresh event loop on a new thread and return its result.
    """
    q: Queue = Queue()

    def _runner():
        result = asyncio.run(coro)
        q.put(result)

    t = threading.Thread(target=_runner, daemon=True)
    t.start()
    res = q.get()  # wait for completion
    t.join()
    return res


@pytest.mark.asyncio
async def test_different_loops_get_different_clients(patch_factory_and_reset):
    # Client on current loop
    c_main = await loop_redis.get_async_redis_connection_in_eventloop()

    # Client on a separate loop/thread
    c_other = run_coro_in_new_loop(loop_redis.get_async_redis_connection_in_eventloop())

    assert c_main is not c_other
    assert c_main.loop is asyncio.get_running_loop()
    # c_other.loop is from a different thread/loop
    assert c_other.loop is not asyncio.get_running_loop()

    # Two factories called (one per loop)
    assert patch_factory_and_reset["n"] == 2


@pytest.mark.asyncio
async def test_close_creates_new_client_afterward(patch_factory_and_reset):
    c1 = await loop_redis.get_async_redis_connection_in_eventloop()
    await loop_redis.close_loop_redis()
    # After close, getting again should create a fresh redis
    c2 = await loop_redis.get_async_redis_connection_in_eventloop()

    assert c1 is not c2
    assert isinstance(c1, FakeClient) and isinstance(c2, FakeClient)
    assert c1.closed is True
    assert c1.connection_pool.disconnected is True

    # two constructions on this loop (before and after close)
    assert patch_factory_and_reset["n"] == 2



import os

from dotenv import load_dotenv

from rediskit.utils import base64_json_to_dict

if os.path.isfile("private.env"):
    load_dotenv("private.env")
if os.path.isfile(".env"):
    load_dotenv(".env")

# Redis Settings
REDIS_HOST = os.environ.get("REDIS_HOST", "localhost")
REDIS_PORT = int(os.environ.get("REDIS_PORT", "6379"))
REDIS_PASSWORD = os.environ.get("REDIS_PASSWORD", "")
REDIS_TOP_NODE = os.environ.get("REDIS_TOP_NODE", "redis_kit_node")
REDIS_SCAN_COUNT = int(os.environ.get("REDIS_SCAN_COUNT", "10000"))
REDIS_SKIP_CACHING = os.environ.get("REDIS_SKIP_CACHING", "false").upper() == "TRUE"

# Lock Settings
REDIS_KIT_LOCK_SETTINGS_REDIS_NAMESPACE = os.environ.get("REDIS_KIT_LOCK_SETTINGS_REDIS_NAMESPACE", f"{REDIS_TOP_NODE}:LOCK")
REDIS_KIT_LOCK_ASYNC_SETTINGS_REDIS_NAMESPACE = os.environ.get("REDIS_KIT_LOCK_ASYNC_SETTINGS_REDIS_NAMESPACE", f"{REDIS_TOP_NODE}:LOCK_ASYNC")
REDIS_KIT_LOCK_CACHE_MUTEX = os.environ.get("REDIS_KIT_LOCK_CACHE_MUTEX", "REDIS_KIT_LOCK_CACHE_MUTEX")

# Encryption Settings
REDIS_KIT_ENCRYPTION_SECRET = base64_json_to_dict(os.environ.get("REDIS_KIT_ENCRYPTION_SECRET", ""))

# Semaphore Settings
REDIS_KIT_SEMAPHORE_SETTINGS_REDIS_NAMESPACE = os.environ.get("REDIS_KIT_SEMAPHORE_SETTINGS_REDIS_NAMESPACE", f"{REDIS_TOP_NODE}:Semaphore")
REDIS_KIT_SEMAPHORE_SETTINGS_STALE_TIMEOUT_SECONDS = int(os.environ.get("REDIS_KIT_SEMAPHORE_SETTINGS_STALE_TIMEOUT_SECONDS", "30"))
REDIS_KIT_SEMAPHORE_LOCK_TIME_TO_LIVE = int(os.environ.get("REDIS_KIT_SEMAPHORE_LOCK_TIME_TO_LIVE", "30"))



# Re-export the two client modules under tidy names
from rediskit.redis import a_client, client
from rediskit.redis.node import get_redis_top_node

__all__ = ["client", "a_client", "get_redis_top_node"]



import json
from typing import Any, Callable, Mapping

from redis import asyncio as redis_async

from rediskit.encrypter import Encrypter
from rediskit.redis.a_client.connection import get_async_redis_connection
from rediskit.redis.node import get_redis_top_node


async def hash_set_ttl_for_key(
    tenant_id: str | None,
    key: str,
    fields: list[str],
    ttl: int,
    connection: redis_async.Redis | None = None,
    top_node: Callable[..., str] = get_redis_top_node,
) -> None:
    node_key = top_node(tenant_id, key)
    conn = connection if connection is not None else get_async_redis_connection()
    await conn.hexpire(node_key, ttl, *fields)  # type: ignore[attr-defined]


async def h_set_cache_to_redis(
    tenant_id: str | None,
    key: str | None,
    fields: dict[str, Any],
    top_node: Callable = get_redis_top_node,
    connection: redis_async.Redis | None = None,
    ttl: int | None = None,
    enable_encryption: bool = False,
) -> None:
    node_key = top_node(tenant_id, key)
    conn = connection if connection is not None else get_async_redis_connection()
    if enable_encryption:
        mapping: Mapping[str | bytes, str | bytes | int | float] = {
            field: Encrypter().encrypt(json.dumps(value).encode("utf-8")) for field, value in fields.items()
        }
    else:
        mapping = {field: json.dumps(value) for field, value in fields.items()}
    await conn.hset(node_key, mapping=mapping)
    if ttl is not None:
        await conn.hexpire(node_key, ttl, *mapping.keys())  # type: ignore[attr-defined]


async def h_get_cache_from_redis(
    tenant_id: str | None,
    key: str | None,
    fields: str | list[str] | None = None,
    top_node: Callable = get_redis_top_node,
    connection: redis_async.Redis | None = None,
    set_ttl_on_read: int | None = None,
    is_encrypted: bool = False,
) -> dict[str, Any] | None:
    node_key = top_node(tenant_id, key)
    conn = connection if connection is not None else get_async_redis_connection()

    if fields is None:
        result = await conn.hgetall(node_key)
        data = {field: value for field, value in result.items()}
    elif isinstance(fields, str):
        value = await conn.hget(node_key, fields)
        data = {fields: (value if value is not None else None)}
    elif isinstance(fields, list):
        values = await conn.hmget(node_key, fields)
        data = {fields[i]: (value if value is not None else None) for i, value in enumerate(values)}
    else:
        raise ValueError("fields must be either None, a string, or a list of strings")

    if set_ttl_on_read is not None and data:
        await conn.hexpire(node_key, set_ttl_on_read, *data.keys())  # type: ignore[attr-defined]

    if is_encrypted:
        result = {k: json.loads(Encrypter().decrypt(v)) for k, v in data.items() if v is not None}
    else:
        result = {k: json.loads(v) for k, v in data.items() if v is not None}

    return result


async def h_scan_fields(
    tenant_id: str | None,
    key: str | None,
    match: str,
    top_node: Callable = get_redis_top_node,
    connection: redis_async.Redis | None = None,
) -> list[str]:
    node_key = top_node(tenant_id, key)
    conn = connection if connection is not None else get_async_redis_connection()

    matched: list[str] = []
    async for field, value in conn.hscan_iter(node_key, match=match):
        matched.append(field)
    return matched


async def h_del_cache_from_redis(
    tenant_id: str | None,
    key: str | None,
    fields: dict[str, Any] | list[str],
    top_node: Callable = get_redis_top_node,
    connection: redis_async.Redis | None = None,
) -> None:
    node_key = top_node(tenant_id, key)
    conn = connection if connection is not None else get_async_redis_connection()
    if isinstance(fields, dict):
        field_names = list(fields.keys())
    elif isinstance(fields, list):
        field_names = fields
    else:
        raise ValueError("fields must be either a dictionary or a list of strings")
    await conn.hdel(node_key, *field_names)



from typing import Callable

from redis import asyncio as redis_async

from rediskit.redis.a_client.connection import get_async_redis_connection
from rediskit.redis.a_client.keys_op import set_ttl_for_key
from rediskit.redis.client.connection import log
from rediskit.redis.node import get_redis_top_node


async def load_blob_from_redis(
    tenant_id: str | None,
    match: str | None,
    connection: redis_async.Redis | None = None,
    set_ttl_on_read: int | None = None,
) -> bytes | None:
    log.info(f"Loading cache from redis tenantId:{tenant_id}, key: {match}")
    conn = connection if connection is not None else get_async_redis_connection()
    node_match = get_redis_top_node(tenant_id, match)
    encoded = await conn.get(node_match)
    if encoded is None:
        return None
    if set_ttl_on_read:
        await set_ttl_for_key(tenant_id, match, ttl=set_ttl_on_read)
    return encoded


async def dump_blob_to_redis(
    tenant_id: str | None,
    key: str | None,
    payload: str,
    top_node: Callable = get_redis_top_node,
    connection: redis_async.Redis | None = None,
    ttl: int | None = None,
) -> None:
    log.info(f"Dump cache tenantId:{tenant_id}, key: {key}")
    node_key = top_node(tenant_id, key)
    conn = connection if connection is not None else get_async_redis_connection()
    await conn.set(node_key, payload)
    if ttl is not None:
        await conn.expire(node_key, ttl)



import asyncio
import functools
import random
from dataclasses import asdict, dataclass
from typing import Callable, ParamSpec, TypeVar

import httpx

P = ParamSpec("P")
R = TypeVar("R")


@dataclass(frozen=True)
class RetryPolicy:
    attempts: int = 3
    backoff: float = 0.5  # initial delay (seconds)
    jitter: float = 0.1  # +/- seconds (fixed) — see note below
    exceptions: tuple[type[BaseException], ...] = (httpx.RequestError, httpx.HTTPStatusError)
    enabled: bool = True
    max_backoff: float | None = None  # optional cap


def _normalize_policy(default: RetryPolicy, override: "RetryPolicy | dict | None") -> RetryPolicy:
    if override is None:
        return default
    if isinstance(override, RetryPolicy):
        return override
    if isinstance(override, dict):
        base = asdict(default)
        base.update(override)
        # normalize exceptions if provided as a single class or iterable
        excs = base.get("exceptions")
        if excs is not None and not isinstance(excs, tuple):
            if isinstance(excs, type) and issubclass(excs, BaseException):
                base["exceptions"] = (excs,)
            else:
                base["exceptions"] = tuple(excs)
        return RetryPolicy(**base)
    return default


def retry_async(default: RetryPolicy | None = None) -> Callable[[Callable[P, R]], Callable[P, R]]:
    default = default or RetryPolicy()

    def deco(fn: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(fn)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            policy = _normalize_policy(default, kwargs.pop("retry_policy", None))
            if not policy.enabled or policy.attempts <= 1:
                return await fn(*args, **kwargs)

            delay = max(0.0, policy.backoff)
            for attempt in range(1, policy.attempts + 1):
                try:
                    return await fn(*args, **kwargs)
                except policy.exceptions:
                    if attempt == policy.attempts:
                        raise
                    # fixed jitter around delay; consider decorrelated/full jitter below
                    j = random.uniform(-policy.jitter, policy.jitter) if policy.jitter else 0.0
                    sleep_for = max(0.0, delay + j)
                    await asyncio.sleep(sleep_for)
                    delay = min(delay * 2, policy.max_backoff) if policy.max_backoff else delay * 2

        return wrapper

    return deco



import base64
import inspect
import json
import pickle
from typing import Any, Callable, Literal

import zstd

from rediskit import Encrypter, config

cache_type_options = Literal["zipPickled", "zipJson"]
redis_storage_options = Literal["string", "hash"]


def split_hash_key(key: str) -> tuple[str, str]:
    *parts, field = key.split(":")
    if not parts:
        raise ValueError("Cannot use a single-part key with hash storage.")
    return ":".join(parts), field


def compress_and_sign(data: Any, serialize_fn: Callable[[Any], bytes], enable_encryption: bool = False) -> str:
    serialized_data = serialize_fn(data)
    if enable_encryption:
        compressed_data = Encrypter().encrypt(serialized_data)
    else:
        compressed_data = zstd.compress(serialized_data)

    return base64.b64encode(compressed_data).decode("utf-8")


def verify_and_decompress(payload: bytes, deserialize_fn: Callable[[bytes], Any], enable_encryption: bool = False) -> Any:
    if enable_encryption:
        serialized_data = Encrypter().decrypt(payload)
    else:
        serialized_data = zstd.decompress(payload)
    return deserialize_fn(serialized_data)


def deserialize_data(
    data: Any,
    cache_type: cache_type_options,
    enable_encryption: bool = False,
) -> bytes:
    if cache_type == "zipPickled":
        cached_data = verify_and_decompress(base64.b64decode(data), lambda b: pickle.loads(b), enable_encryption)
    elif cache_type == "zipJson":
        cached_data = verify_and_decompress(base64.b64decode(data), lambda b: json.loads(b.decode("utf-8")), enable_encryption)
    else:
        raise ValueError("Unknown cacheType specified.")

    return cached_data


def serialize_data(
    data: Any,
    cache_type: cache_type_options,
    enable_encryption: bool = False,
) -> str:
    if cache_type == "zipPickled":
        payload = compress_and_sign(data, lambda d: pickle.dumps(d), enable_encryption)
    elif cache_type == "zipJson":
        payload = compress_and_sign(data, lambda d: json.dumps(d).encode("utf-8"), enable_encryption)
    else:
        raise ValueError("Unknown cacheType specified.")
    return payload


def compute_value[T](param: T | Callable[..., T], *args, **kwargs) -> T:
    if callable(param):
        sig = inspect.signature(param)
        accepts_kwargs = any(p.kind == inspect.Parameter.VAR_KEYWORD for p in sig.parameters.values())

        if accepts_kwargs:
            # Pass all kwargs directly
            value = param(*args, **kwargs)
        else:
            # Filter only matching kwargs
            filtered_kwargs = {k: v for k, v in kwargs.items() if k in sig.parameters}
            bound = sig.bind(*args, **filtered_kwargs)
            bound.apply_defaults()
            value = param(*bound.args, **bound.kwargs)
        return value
    else:
        return param


def get_params(
    func: Callable, memoize_key: Callable[..., str] | str, ttl: Callable[..., int] | int | None, bypass_cache: Callable[..., bool] | bool, *args, **kwargs
) -> tuple:
    def compute_memoize_key(*args, **kwargs) -> str:
        if not (isinstance(memoize_key, str) or callable(memoize_key)):
            raise ValueError(f"Expected memoizeKey to be Callable or a str. got {type(memoize_key)}")
        return compute_value(memoize_key, *args, **kwargs)

    def compute_ttl(*args, **kwargs) -> int | None:
        if ttl is None:
            return None
        if not (isinstance(ttl, int) or callable(ttl)):
            raise ValueError(f"Expected ttl to be Callable or an int. got {type(ttl)}")
        return compute_value(ttl, *args, **kwargs)

    def compute_by_pass_cache(*args, **kwargs) -> bool:
        if not (isinstance(bypass_cache, bool) or callable(bypass_cache)):
            raise ValueError(f"Expected bypassCache to be Callable or an int. got {type(bypass_cache)}")
        return compute_value(bypass_cache, *args, **kwargs)

    def compute_tenant_id(wrapped_func: Callable[..., Any], *args, **kwargs) -> str | None:
        bound_args = inspect.signature(wrapped_func).bind(*args, **kwargs)
        bound_args.apply_defaults()
        tenant_id = bound_args.arguments.get("tenantId") or bound_args.kwargs.get("tenantId")
        return tenant_id

    def get_lock_name(tenant_id: str | None, computed_memoize_key: str) -> str:
        if tenant_id is None:
            return f"{config.REDIS_KIT_LOCK_CACHE_MUTEX}:{computed_memoize_key}"
        else:
            return f"{config.REDIS_KIT_LOCK_CACHE_MUTEX}:{tenant_id}:{computed_memoize_key}"

    def params_calc(func, *args, **kwargs) -> tuple[str, int | None, str | None, str, bool]:
        computed_memoize_key = compute_memoize_key(*args, **kwargs)
        computed_ttl = compute_ttl(*args, **kwargs)
        tenant_id = compute_tenant_id(func, *args, **kwargs)
        lock_name = get_lock_name(tenant_id, computed_memoize_key)
        by_pass_cached_data = compute_by_pass_cache(*args, **kwargs)

        return computed_memoize_key, computed_ttl, tenant_id, lock_name, by_pass_cached_data

    return params_calc(func, *args, **kwargs)



import base64
import datetime
import enum
import json
import logging
import uuid
from typing import Any


def base64_json_to_dict(keys_base64: str | None) -> dict[str, str]:
    if not keys_base64:
        logging.error("No keys_base64 provided")
    try:
        # Decode from base64 to a JSON string, then load it into a dict
        decodedJson = base64.b64decode(keys_base64).decode("utf-8")
        decoded = json.loads(decodedJson)
        return decoded
    except Exception:
        logging.error("No keys_base64 provided")


def json_encoder(value: Any, raise_if_no_match: bool = False):
    if isinstance(value, enum.Enum):
        return value.value
    elif isinstance(value, uuid.UUID):
        return str(value)
    elif isinstance(value, (datetime.datetime, datetime.date)):
        return value.isoformat()
    elif raise_if_no_match:
        raise TypeError(f"Object of type {value.__class__.__name__} is not JSON serializable")
    return value


def serialize_values(value: Any) -> Any:
    if isinstance(value, dict):
        serialized_dict = {}
        for dictKey, dictValue in value.items():
            serialized_key = json_encoder(dictKey)
            serialized_dict[serialized_key] = serialize_values(dictValue)
        return serialized_dict
    elif isinstance(value, list):
        serialized_list = [serialize_values(v) for v in value]
        return serialized_list
    return json_encoder(value)


def dict_to_list(items: dict | None) -> list[Any]:
    list_items: list[Any] = []
    if items is None:
        return list_items
    for key in items:
        list_items.append(items[key])
    return list_items


def deserialize_dict_model_property(items: dict | None, model_type: Any) -> None:
    if isinstance(items, dict):
        for key in items:
            value = items[key]
            if isinstance(value, dict):
                items[key] = model_type(**value)


def merge_dict_data(old_dict: dict, new_dict: dict) -> dict:
    merged_dict = dict(old_dict)
    for key in new_dict:
        if key in merged_dict:
            if isinstance(merged_dict[key], dict) and isinstance(new_dict[key], dict):
                merged_dict[key] = merge_dict_data(merged_dict[key], new_dict[key])
            else:
                merged_dict[key] = new_dict[key]
        else:
            merged_dict[key] = new_dict[key]
    return merged_dict


def remove_nones_from_dict_data(original_dict: dict) -> dict:
    clean_dict = dict(original_dict)
    keys_to_pop = []
    for key in clean_dict:
        if isinstance(clean_dict[key], dict):
            clean_dict[key] = remove_nones_from_dict_data(clean_dict[key])
        else:
            if clean_dict[key] is None:
                keys_to_pop.append(key)
    for key_to_pop in keys_to_pop:
        clean_dict.pop(key_to_pop)
    return clean_dict


def remove_matching_dict_data(original_dict: dict, matching_dict: dict) -> tuple[dict, dict]:
    changed_data = {}
    clean_dict = dict(matching_dict)
    keys_to_pop = []
    for key in clean_dict:
        if key in original_dict:
            if isinstance(original_dict[key], dict) and isinstance(clean_dict[key], dict):
                data = remove_matching_dict_data(original_dict[key], clean_dict[key])
                clean_dict[key] = data[0]
                changed_data[key] = data[1]
            elif isinstance(original_dict[key], list) and isinstance(matching_dict[key], list):
                if check_matching_list_data(original_dict[key], matching_dict[key]):
                    keys_to_pop.append(key)
                else:
                    changed_data[key] = original_dict[key]
            else:
                if original_dict[key] == clean_dict[key]:
                    keys_to_pop.append(key)
                else:
                    changed_data[key] = original_dict[key]
    for key_to_pop in keys_to_pop:
        clean_dict.pop(key_to_pop)
    return clean_dict, changed_data


def check_matching_dict_data(original_dict: dict, matching_dict: dict) -> bool:
    matching = True
    for key in matching_dict:
        if key in original_dict:
            if isinstance(original_dict[key], dict) and isinstance(matching_dict[key], dict):
                matching = check_matching_dict_data(original_dict[key], matching_dict[key])
            elif isinstance(original_dict[key], list) and isinstance(matching_dict[key], list):
                matching = check_matching_list_data(original_dict[key], matching_dict[key])
            else:
                matching = original_dict[key] == matching_dict[key]
        else:
            matching = False
        if not matching:
            break
    return matching


def check_empty_dict_data(data: dict) -> bool:
    empty = True
    for key in data:
        if isinstance(data[key], dict):
            empty = check_empty_dict_data(data[key])
        else:
            empty = False
        if not empty:
            break
    return empty


def check_matching_list_data(original_list: list, matching_list: list) -> bool:
    matching = len(original_list) == len(matching_list)
    for i in range(len(original_list)):
        if not matching:
            break
        if isinstance(original_list[i], dict) and isinstance(matching_list[i], dict):
            matching = check_matching_dict_data(original_list[i], matching_list[i])
        elif isinstance(original_list[i], list) and isinstance(matching_list[i], list):
            matching = check_matching_list_data(original_list[i], matching_list[i])
        else:
            matching = original_list[i] == matching_list[i]
    return matching


def remove_keys(data: dict, key_map: dict, ignore_keys: list[str] | None = None) -> None:
    if ignore_keys is None:
        ignore_keys = ["id"]
    for key in key_map:
        if key in data and key not in ignore_keys:
            if isinstance(key_map[key], dict):
                if isinstance(data[key], dict):
                    remove_keys(data[key], key_map[key])
            else:
                data.pop(key)



import asyncio
import time

import pytest

from rediskit.memoize import a_redis_memoize
from rediskit.redis.a_client import init_async_redis_connection_pool

TEST_TENANT_ID = "TEST_TENANT_REDIS_CACHE"


@pytest.mark.asyncio
async def testAsyncCaching():
    await init_async_redis_connection_pool()

    @a_redis_memoize(memoize_key="testAsync", ttl=10, cache_type="zipJson")
    async def slowFunc(tenantId: str, x):
        await asyncio.sleep(1)
        return {"asyncResult": x}

    x = 777
    start = time.time()
    res1 = await slowFunc(TEST_TENANT_ID, x)
    duration1 = time.time() - start

    start = time.time()
    res2 = await slowFunc(TEST_TENANT_ID, x)
    duration2 = time.time() - start

    assert res1.get("asyncResult") == x
    assert res2.get("asyncResult") == x
    assert duration1 >= 1.0
    assert duration2 < 0.5


@pytest.mark.asyncio
async def testAsyncBypass():
    await init_async_redis_connection_pool()

    def bypassFunc(*args, **kwargs):
        return kwargs.get("forceBypass", False)

    @a_redis_memoize(memoize_key="testAsyncBypass", ttl=10, cache_type="zipJson", bypass_cache=bypassFunc)
    async def slowFunc(tenantId: str, x, forceBypass=False):
        await asyncio.sleep(1)
        return {"asyncResult": x}

    x = 888
    start = time.time()
    res1 = await slowFunc(TEST_TENANT_ID, x, forceBypass=False)
    duration1 = time.time() - start

    start = time.time()
    res2 = await slowFunc(TEST_TENANT_ID, x, forceBypass=False)
    duration2 = time.time() - start

    start = time.time()
    res3 = await slowFunc(TEST_TENANT_ID, x, forceBypass=True)
    duration3 = time.time() - start

    assert duration1 >= 1.0
    assert duration2 < 0.5
    assert duration3 >= 1.0
    assert res1.get("asyncResult") == x
    assert res2.get("asyncResult") == x
    assert res3.get("asyncResult") == x


@pytest.mark.asyncio
async def testReturnNoneAsync():
    await init_async_redis_connection_pool()

    @a_redis_memoize(memoize_key="returnNoneAsync", ttl=10, cache_type="zipJson")
    async def func(tenantId: str, x):
        await asyncio.sleep(1)
        return None

    res1 = await func(TEST_TENANT_ID, 42)
    res2 = await func(TEST_TENANT_ID, 42)
    assert res1 is None
    assert res2 is None


@pytest.mark.asyncio
async def testConcurrentAsyncCaching():
    await init_async_redis_connection_pool()
    # Test that concurrent async calls only execute the function once
    call_count = [0]

    @a_redis_memoize(memoize_key="concurrentAsync", ttl=10, cache_type="zipJson")
    async def func(tenantId: str, x):
        call_count[0] += 1
        await asyncio.sleep(1)
        return {"result": x}

    tasks = [func(TEST_TENANT_ID, 888) for _ in range(5)]
    results = await asyncio.gather(*tasks)
    for r in results:
        assert r.get("result") == 888
    # The function should have been executed only once
    assert call_count[0] == 1



"""
rediskit - Redis-backed performance and concurrency primitives for Python applications.

Provides caching, distributed coordination, and data protection using Redis.
"""

from rediskit.async_semaphore import AsyncSemaphore
from rediskit.encrypter import Encrypter
from rediskit.memoize import a_redis_memoize, redis_memoize
from rediskit.redis import a_client, client, get_redis_top_node
from rediskit.redis_lock import get_async_redis_mutex_lock, get_redis_mutex_lock
from rediskit.retry_decorator import RetryPolicy, retry_async
from rediskit.semaphore import Semaphore

__all__ = [
    # Redis client
    "get_redis_top_node",
    "a_client",
    "client",
    # Redis
    "redis_memoize",
    "a_redis_memoize",
    "get_redis_mutex_lock",
    "get_async_redis_mutex_lock",
    "Semaphore",
    "AsyncSemaphore",
    "RetryPolicy",
    "retry_async",
    # Encryption,
    "Encrypter",
]



from typing import Callable

from redis import Redis

from rediskit.redis.client.connection import get_redis_connection, log
from rediskit.redis.client.keys_op import set_ttl_for_key
from rediskit.redis.node import get_redis_top_node


def load_blob_from_redis(tenant_id: str | None, match: str | None, connection: Redis | None = None, set_ttl_on_read: int | None = None) -> bytes | None:
    log.info(f"Loading cache from redis tenantId:{tenant_id}, key: {match}")
    connection = connection if connection is not None else get_redis_connection()
    node_match = get_redis_top_node(tenant_id, match)
    # Retrieve raw bytes directly from Redis.
    encoded = connection.get(node_match)
    if encoded is None:
        return None
    if set_ttl_on_read:
        set_ttl_for_key(tenant_id, match, ttl=set_ttl_on_read)

    return encoded


def dump_blob_to_redis(
    tenant_id: str | None, key: str | None, payload: str, top_node: Callable = get_redis_top_node, connection: Redis | None = None, ttl: int | None = None
) -> None:
    log.info(f"Dump cache tenantId:{tenant_id}, key: {key}")
    node_key = top_node(tenant_id, key)
    connection = connection if connection is not None else get_redis_connection()
    connection.set(node_key, payload)
    if ttl is not None:
        connection.expire(node_key, ttl)



from contextlib import asynccontextmanager

from redis import asyncio as redis_async

from rediskit import config
from rediskit.redis.a_client.redis_in_eventloop import close_loop_redis, get_async_client_for_current_loop, get_async_redis_connection_in_eventloop


async def init_async_redis_connection_pool() -> None:
    await get_async_redis_connection_in_eventloop()


@asynccontextmanager
async def redis_single_connection_context():
    pool = redis_async.ConnectionPool(
        host=config.REDIS_HOST,
        port=config.REDIS_PORT,
        password=config.REDIS_PASSWORD,
        decode_responses=True,
        max_connections=1,
    )
    client = redis_async.Redis(connection_pool=pool)
    try:
        yield client
    finally:
        await client.aclose()
        await pool.disconnect()


def get_async_redis_connection() -> redis_async.Redis:
    return get_async_client_for_current_loop()


async def async_connection_close() -> None:
    await close_loop_redis()



import pytest

from rediskit.redis import get_redis_top_node
from rediskit.redis.client import get_redis_connection, h_del_cache_from_redis, h_get_cache_from_redis, h_set_cache_to_redis, list_keys

TEST_TENANT_ID = "TEST_HASH_TENANT_REDIS"


@pytest.fixture
def Connection():
    return get_redis_connection()


@pytest.fixture(autouse=True)
def CleanupRedis(Connection):
    prefix = get_redis_top_node(TEST_TENANT_ID, "")
    for key in Connection.scan_iter(match=f"{prefix}*"):
        Connection.delete(key)
    yield
    for key in Connection.scan_iter(match=f"{prefix}*"):
        Connection.delete(key)


def test_HSetCacheToRedis_and_HGetCacheFromRedis_plain(Connection):
    node_key = "mytest"
    fields = {"a": 1, "b": "hello"}
    h_set_cache_to_redis(TEST_TENANT_ID, node_key, fields, connection=Connection)
    result = h_get_cache_from_redis(TEST_TENANT_ID, node_key, None, connection=Connection)
    assert result == fields


def test_HSetCacheToRedis_with_ttl(Connection):
    node_key = "ttltest"
    fields = {"x": 99}
    h_set_cache_to_redis(TEST_TENANT_ID, node_key, fields, connection=Connection, ttl=30)
    result = h_get_cache_from_redis(TEST_TENANT_ID, node_key, list(fields.keys()), connection=Connection)
    assert result == fields


def test_HSetCacheToRedis_and_HGetCacheFromRedis_encrypted(Connection):
    node_key = "cryptotest"
    fields = {"foo": [1, 2, 3], "bar": "baz"}
    h_set_cache_to_redis(TEST_TENANT_ID, node_key, fields, connection=Connection, enable_encryption=True)
    # By default, must specify isEncrypted
    result = h_get_cache_from_redis(TEST_TENANT_ID, node_key, list(fields.keys()), connection=Connection, is_encrypted=True)
    assert result == fields


def test_HDelCacheFromRedis_dict_and_list(Connection):
    node_key = "deltest"
    fields = {"a": 111, "b": 222, "c": 333}
    h_set_cache_to_redis(TEST_TENANT_ID, node_key, fields, connection=Connection)

    # Delete one by dict
    h_del_cache_from_redis(TEST_TENANT_ID, node_key, {"a": fields["a"]}, connection=Connection)
    state = h_get_cache_from_redis(TEST_TENANT_ID, node_key, None, connection=Connection)
    assert set(state.keys()) == {"b", "c"}

    # Delete rest by list
    h_del_cache_from_redis(TEST_TENANT_ID, node_key, ["b", "c"], connection=Connection)
    state = h_get_cache_from_redis(TEST_TENANT_ID, node_key, None, connection=Connection)
    assert state == {}


def test_HGetCacheFromRedis_fields_variants(Connection):
    node_key = "multitype"
    fields = {"s": "value", "n": 42}
    h_set_cache_to_redis(TEST_TENANT_ID, node_key, fields, connection=Connection)
    # Get single field (str)
    res = h_get_cache_from_redis(TEST_TENANT_ID, node_key, "s", connection=Connection)
    assert res == {"s": "value"}
    # Get multiple (list)
    res2 = h_get_cache_from_redis(TEST_TENANT_ID, node_key, ["s", "n"], connection=Connection)
    assert res2 == fields
    # All (None)
    res3 = h_get_cache_from_redis(TEST_TENANT_ID, node_key, None, connection=Connection)
    assert res3 == fields


def test_HGetCacheFromRedis_invalid_field_type(Connection):
    node_key = "badtype"
    h_set_cache_to_redis(TEST_TENANT_ID, node_key, {"f": 1}, connection=Connection)
    with pytest.raises(ValueError):
        h_get_cache_from_redis(TEST_TENANT_ID, node_key, 42, connection=Connection)


def test_HDelCacheFromRedis_invalid_field_type(Connection):
    node_key = "delfail"
    h_set_cache_to_redis(TEST_TENANT_ID, node_key, {"f": 1}, connection=Connection)
    with pytest.raises(ValueError):
        h_del_cache_from_redis(TEST_TENANT_ID, node_key, 1.23, connection=Connection)


def test_ListKeys_yields_all(Connection):
    # Populate multiple keys
    for i in range(3):
        h_set_cache_to_redis(TEST_TENANT_ID, f"key{i}", {"val": i}, connection=Connection)
    # All keys for this "directory"
    result = list(list_keys(TEST_TENANT_ID, "*", connection=Connection))
    assert len(result) >= 3  # At least 3, may have more if other tests collide


def test_ListKeys_limit_over_10000_raises(Connection):
    # Insert 10,001 keys
    for i in range(10_001):
        h_set_cache_to_redis(TEST_TENANT_ID, f"massive{i}", {"x": i}, connection=Connection)
    with pytest.raises(ValueError, match="exceeded 10_000"):
        list(list_keys(TEST_TENANT_ID, "massive*", connection=Connection))
